# -*- coding: utf-8 -*-
"""# pyncm
本模块提供`Get/Set/DumpCurrentSession`,`LoadNewSessionFromDump` 以管理请求
API 使用请参见 `pyncm.apis` 文档 
"""
from typing import Text, Union
from time import time
from .utils.crypto import RandomString, EapiEncrypt, EapiDecrypt, HexCompose
import requests, logging, json

__version__ = "1.6.5.7"


class Session(requests.Session):
    """Represents an API session"""

    HOST = "music.163.com"
    UA_DEFAULT = (
        "Mozilla/5.0 (linux@github.com/greats3an/pyncm) Chrome/PyNCM.%s" % __version__
    )
    UA_EAPI = "NeteaseMusic/7.2.24.1597753235(7002024);Dalvik/2.1.0 (Linux; U; Android 11; Pixel 2 XL Build/RP1A.200720.009)"
    UA_LINUX_API = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36"
    CONFIG_EAPI = {
        "appver": "9.9.99",
        "buildver": "9009099",
        "channel": "offical",
        "deviceId": RandomString(8),
        "mobilename": "Pixel2XL",
        "os": "android",
        "osver": "10.1",
        "resolution": "2712x1440",
        "versioncode": "240",
    }
    force_http = False

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "User-Agent": self.UA_DEFAULT,
            "Referer": self.HOST,
        }
        self.login_info = {"success": False, "tick": time(), "content": None}
        self.csrf_token = ""

    # region Shorthands
    @property
    def uid(self):
        """用户 ID"""
        return self.login_info["content"]["account"]["id"] if self.logged_in else 0

    @property
    def nickname(self):
        """登陆用户的昵称"""
        return (
            self.login_info["content"]["profile"]["nickname"] if self.logged_in else ""
        )

    @property
    def lastIP(self):
        """登陆时，上一次登陆的 IP"""
        return (
            self.login_info["content"]["profile"]["lastLoginIP"]
            if self.logged_in
            else ""
        )

    @property
    def vipType(self):
        """账号 VIP 等级"""
        return self.login_info["content"]["profile"]["vipType"] if self.logged_in else 0

    @property
    def logged_in(self):
        """是否已经登陆"""
        return self.login_info["success"]

    # endregion
    def request(
        self, method: str, url: Union[str, bytes, Text], *a, **k
    ) -> requests.Response:
        """Initiates & fires a request

        Args:
            method (str): HTTP Verb
            url (Union[str, bytes, Text]): Full / partial API URL

        Returns:
            requests.Response: A requests.Response object
        """
        if url[:4] != "http":
            url = 'https://%s%s' % (self.HOST,url)         
        if self.force_http:
            url = url.replace('https:','http:')            
        return super().request(method, url, *a, **k)

    # region symbols for loading/reloading authentication info
    _session_info = {        
        "login_info": (
            lambda self: getattr(self, "login_info"),
            lambda self, v: setattr(self, "login_info", v),
        ),
        "csrf_token": (
            lambda self: getattr(self, "csrf_token"),
            lambda self, v: setattr(self, "csrf_token", v),
        ),
        "cookies": (
            lambda self: [
                {"name": c.name, "value": c.value, "domain": c.domain, "path": c.path}
                for c in getattr(self, "cookies")
            ],
            lambda self, cookies: [
                getattr(self, "cookies").set(**cookie) for cookie in cookies
            ],
        ),
    }

    def dump(self) -> dict:
        """Dumps current session info as dictionary"""
        return {
            name: self._session_info[name][0](self)
            for name in self._session_info.keys()
        }

    def load(self, dumped):
        """Loads session info from dictionary"""
        for k, v in dumped.items():
            self._session_info[k][1](self, v)
        return True

    # endregion


class SessionManager:
    """Manages session switching / loading / dumping"""

    def __init__(self) -> None:
        self.session = Session()

    def get(self):
        return self.session

    def set(self, session):
        self.session = session

    # region Session serialization
    @staticmethod
    def stringify(session: Session) -> str:
        return EapiEncrypt("pyncm", json.dumps(session.dump()))["params"]

    @staticmethod
    def parse(dump: str) -> Session:
        session = Session()
        dump = HexCompose(dump)
        dump = EapiDecrypt(dump).decode()
        dump = dump.split("-36cd479b6b5-")
        assert dump[0] == "pyncm"  # check magic
        session.load(json.loads(dump[1]))  # loading config dict
        return session

    # endregion


logger = logging.getLogger("ncm")
"""Logger to be used by local modules"""
sessionManager = SessionManager()
"""Manages active session"""


def GetCurrentSession() -> Session:
    """Retrives current active session"""
    return sessionManager.get()


def SetCurrentSession(session: Session):
    """Sets current active session"""
    sessionManager.set(session)


def SetNewSession():
    """Creates and sets new session"""
    sessionManager.set(Session())


def LoadSessionFromString(dump: str) -> Session:
    """Loads a session from dumped string"""
    session = SessionManager.parse(dump)
    return session


def DumpSessionAsString(session: Session) -> str:
    """Dumps session as encrypted string"""
    return SessionManager.stringify(session)



import importlib


def assert_dep(name):
    assert importlib.util.find_spec(name), "需要安装 %s" % name


def login():
    import inquirer
    import 手机登录, 二维码登录

    methods = {"手机登录": 手机登录, "二维码登录": 二维码登录}
    assert methods[
        inquirer.prompt([inquirer.List("method", message="登陆方法", choices=methods)])[
            "method"
        ]
    ].login(), "登录失败"
    from pyncm import GetCurrentSession

    print(
        "已登录",
        GetCurrentSession().nickname,
        "曾用 IP：",
        GetCurrentSession().lastIP,
        "用户 ID：",
        GetCurrentSession().uid,
    )
    return True


assert_dep("pyncm")
assert_dep("inquirer")
from pyncm import __version__

print("PyNCM %s" % __version__)



# -*- coding: utf-8 -*-
"""用户 - User 有关 APIs"""
from . import WeapiCryptoRequest, UserIDBasedApi, LoginRequiredApi


@WeapiCryptoRequest
@UserIDBasedApi
def GetUserDetail(user_id=0):
    """网页端 - 获取某用户资料详情

    Args:
        user_id (int): 用户 ID。置 0 表示当前已登录的用户 . defaults to 0

    Returns:
        dict
    """
    return "/weapi/v1/user/detail/%s" % user_id, {}


@WeapiCryptoRequest
@UserIDBasedApi
def GetUserPlaylists(user_id, offset=0, limit=1001):
    """网页端 - 获取某用户创建的歌单

    Args:
        user_id (int): 用户 ID。置 0 表示当前已登录的用户 . defaults to 0
        offset (int, optional): 获取偏移数. Defaults to 0.
        limit (int, optional): 单次获取量. Defaults to 30.

    Returns:
        dict
    """
    return "/weapi/user/playlist", {
        "offset": str(offset),
        "limit": str(limit),
        "uid": str(user_id),
    }


@WeapiCryptoRequest
@LoginRequiredApi
def GetUserArtistSubs(limit=30):
    """网页端 - 获取收藏歌手内容
    Args:
        limit (int, optional): 单次获取量. Defaults to 30.

    Returns:
        dict
    """
    return "/weapi/artist/sublist", {"limit": str(limit)}


SIGNIN_TYPE_MOBILE = 0
"""移动端签到 +4 EXP"""
SIGNIN_TYPE_WEB = 1
"""网页端签到 +1 EXP"""


@WeapiCryptoRequest
@LoginRequiredApi
def SetSignin(type=0):
    """移动端、PC端 - 每日签到

    Args:
        type (int, optional): 签到类型，请使用本模块内 SIGNIN_TYPE_... 之一 .Defaults to SIGNIN_TYPE_MOBILE.

    Returns:
        dict
    """
    return "/weapi/point/dailyTask", {"type": str(type)}



# -*- coding: utf-8 -*-
"""最嗨电音、爵士电台 - DI.FM APIs"""

from .. import WeapiCryptoRequest


@WeapiCryptoRequest
def GetCurrentPlayingTrackList(channelId=101, limit=10, source=0):
    """移动端 - 小程序 - 获取当前某频道下正在播放的歌曲

    Args:
        channelId (int, optional): 频道 (channel) ID. Defaults to 101.
        limit (int, optional): 单次获取数量. Defaults to 10.
        source (int, optional): [未知]. Defaults to 0.

    Returns:
        dict
    """
    return "/api/dj/difm/playing/tracks/list", {
        "limit": str(limit),
        "source": str(source),
        "channelId": str(channelId),
    }


@WeapiCryptoRequest
def GetChannelCollection(source=0):
    """移动端 - 小程序 - 获取所有频道的信息

    Args:
        source (int, optional): [未知]. Defaults to 0.

    Returns:
        dict
    """
    return "/api/dj/difm/all/style/channel/v2", {"sources": "[%s]" % source}


@WeapiCryptoRequest
def GetChannelSubscriptionCollection(source=0):
    """移动端 - 小程序 - 获取所有*订阅*频道的信息

    Args:
        source (int, optional): [未知]. Defaults to 0.

    Returns:
        dict
    """
    return "/api/dj/difm/subscribe/channels/get/v2", {"sources": "[%s]" % source}


@WeapiCryptoRequest
def SetChannelSubcribiton(id, set_subsubscribe=True):
    """移动端 - 小程序 - 收藏电台

    Args:
        id (str): 电台ID
        set_subsubscribe (bool, optional): 订阅或取消订阅. Defaults to True

    Returns:
        dict
    """
    return (
        "/api/dj/difm/channel/subscribe"
        if set_subsubscribe
        else "/api/dj/difm/channel/unsubscribe",
        {"id": str(id)},
    )



# -*- coding: utf-8 -*-
"""私人FM - Raido APIs"""

from .. import EapiCryptoRequest, LoginRequiredApi


@EapiCryptoRequest
@LoginRequiredApi
def GetMoreRaidoContent(limit=3, e_r=True):
    """PC 端 - 拉取更多FM内容

    Args:
        limit (int, optional): 获取数目. Defaults to 3.
        e_r (bool, optional): 暂存. Defaults to True.

    Returns:
        dict
    """
    return "/api/v1/radio/get", {"limit": str(limit), "e_r": str(e_r).lower()}


@EapiCryptoRequest
@LoginRequiredApi
def SetSkipRadioContent(songId, time=0, alg="itembased", e_r=True):
    """PC 端 - 跳过 FM 歌曲

    Args:
        songId (str): 歌曲ID
        time (int, optional): 播放时长. Defaults to 0.
        alg (str, optional): 暂存. Defaults to "itembased".
        e_r (bool, optional): 暂存. Defaults to True.

    Returns:
        dict
    """
    return "/api/v1/radio/get", {
        "e_r": str(e_r).lower(),
        "songId": str(songId),
        "alg": str(alg),
        "time": str(time),
    }


@EapiCryptoRequest
@LoginRequiredApi
def SetLikeRadioContent(trackId, like=True, time="0", alg="itembased", e_r=True):
    """PC 端 - `收藏喜欢` FM 歌曲

    Args:
        trackId (str): 歌曲ID
        like (bool)：是否收藏
        time (int, optional): 播放时长. Defaults to 0.
        alg (str, optional): 暂存. Defaults to "itembased".
        e_r (bool, optional): 暂存. Defaults to True.

    Returns:
        dict
    """
    return "/api/v1/radio/like", {
        "e_r": str(e_r).lower(),
        "like": str(like).lower(),
        "trackId": str(trackId),
        "alg": str(alg),
        "time": str(time),
    }


@EapiCryptoRequest
@LoginRequiredApi
def SetTrashRadioContent(songId, time="0", alg="itembased", e_r=True):
    """PC 端 - 删除 FM 歌曲

    Args:
        songId (str): 歌曲ID
        time (int, optional): 播放时长. Defaults to 0.
        alg (str, optional): 暂存. Defaults to "itembased".
        e_r (bool, optional): 暂存. Defaults to True.

    Returns:
        dict
    """
    return "/api/v1/radio/trash/add", {
        "e_r": str(e_r).lower(),
        "songId": str(songId),
        "alg": str(alg),
        "time": str(time),
    }



from pyncm.apis.login import LoginViaCellphone
from __init__ import assert_dep


def login():
    import inquirer

    query = inquirer.prompt(
        [
            inquirer.Text("phone", message="手机号"),
            inquirer.Password("password", message="密码"),
            inquirer.Text("ctcode", message="国家代码(默认86)"),
        ]
    )
    return LoginViaCellphone(
        query["phone"], query["password"], ctcode=query["ctcode"] or 86
    )


if __name__ == "__main__":
    print("登录测试")
    assert login(), "登陆失败"



# -*- coding: utf-8 -*-
import random
from hashlib import md5

BASE62 = "PJArHa0dpwhvMNYqKnTbitWfEmosQ9527ZBx46IXUgOzD81VuSFyckLRljG3eC"
BASE64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"


def RandomString(len, chars=BASE62):
    # Generates random string of `len` chars within a selected number of chars
    return "".join([random.choice(chars) for i in range(0, len)])


def HexDigest(data: bytearray):
    # Digests a `bytearray` to a hex string
    return "".join([hex(d)[2:].zfill(2) for d in data])


def HexCompose(hexstr: str):
    # Composes a hex string back to a `bytearray`
    return bytearray([int(hexstr[i : i + 2], 16) for i in range(0, len(hexstr), 2)])


def HashDigest(text):
    # Digests 128 bit md5 hash
    HASH = md5(text.encode("utf-8"))
    return HASH.digest()


def HashHexDigest(text):
    """Digests 128 bit md5 hash,then digest it as a hexstring"""
    return HexDigest(HashDigest(text))



# -*- coding: utf-8 -*-
"""伪装，Anti-AntiCSRF 有关 APIs"""
from . import WeapiCryptoRequest
import json


@WeapiCryptoRequest
def MockWeblog(logs):
    """网页端 - 用户足迹接口 - reserved

    Args:
        logs (Any): 记录的用户足迹

    Returns:
        dict
    """
    return "/weapi/feedback/weblog", {"logs": json.dumps(logs)}



# -*- coding: utf-8 -*-
"""网易云音乐 APIs

## 注意事项
- (#11) 海外用户可能经历 460 "Cheating" 问题，可通过添加以下 Header 解决:    
    GetCurrentSession().headers['X-Real-IP'] = '118.88.88.88'

## 新 API
在获得相关 API 路径和参数后(可参考 Binaryify 大佬的视频 - BV1PY41147r9)
可参考已有 API 函数格式编写相关 PyNCM API. 各API模式修饰器已在本文件注释
"""

from time import time
from requests.models import Response
from ..utils.crypto import (
    EapiDecrypt,
    EapiEncrypt,
    LinuxApiEncrypt,
    WeapiEncrypt,
    AbroadDecrypt,
)
from .. import GetCurrentSession , logger
import json, urllib.parse


class LoginRequiredException(Exception):
    pass


class LoginFailedException(Exception):
    pass


# region Base models export
LOGIN_REQUIRED = LoginRequiredException("Function needs you to be logged in")


def parse(url):
    return urllib.parse.urlparse(url)


def BaseWrapper(requestFunc):
    """Base wrapper for crypto requests"""

    def apiWrapper(apiFunc):
        """apiFunc -> rtype : (url,plain,[method])"""

        def wrapper(*a, **k):
            ret = apiFunc(*a, **k)
            url, plain = ret[:2]
            method = ret[-1] if ret[-1] in ["POST", "GET"] else "POST"
            rsp = requestFunc(url, plain, method)
            try:
                payload = rsp.text if isinstance(rsp, Response) else rsp
                payload = payload.decode() if not isinstance(payload, str) else payload
                payload = json.loads(
                    payload.strip("\x10")
                )  # plaintext responses are also padded...
                if "abroad" in payload and payload["abroad"]:  # addresses #15
                    real_payload = AbroadDecrypt(payload["result"])
                    payload = json.loads(real_payload)
                return payload
            except json.JSONDecodeError:
                return rsp

        return wrapper

    return apiWrapper


def LoginRequiredApi(func):
    """API 需要事先登录"""

    def wrapper(*a, **k):
        if not GetCurrentSession().login_info["success"]:
            raise LOGIN_REQUIRED
        return func(*a, **k)

    return wrapper


def UserIDBasedApi(func):
    """API 第一参数为用户 ID，而该参数可留 0 而指代已登录的用户 ID"""

    def wrapper(user_id=0, *a, **k):
        if user_id == 0 and GetCurrentSession().login_info["success"]:
            user_id = GetCurrentSession().uid
        elif user_id == 0:
            raise LOGIN_REQUIRED
        return func(user_id, *a, **k)

    return wrapper


def EapiEncipered(func):
    """函数值有 Eapi 加密 - 解密并返回原文"""

    def wrapper(*a, **k):
        payload = func(*a, **k)
        try:
            return EapiDecrypt(payload).decode()
        except:
            return payload

    return wrapper


@BaseWrapper
def WeapiCryptoRequest(url, plain, method):
    """Weapi - 适用于 网页端、小程序、手机端部分 APIs"""
    payload = json.dumps({**plain, "csrf_token": GetCurrentSession().csrf_token})
    return GetCurrentSession().request(
        method,
        url.replace("/api/", "/weapi/"),
        params={"csrf_token": GetCurrentSession().csrf_token},
        data={**WeapiEncrypt(payload)},
    )


# region Port of `Binaryify/NeteaseCloudMusicApi`
@BaseWrapper
def LapiCryptoRequest(url, plain, method):
    """Linux API - 适用于 Linux 客户端部分 APIs"""
    payload = {"method": method, "url": GetCurrentSession().HOST + url, "params": plain}
    payload = json.dumps(payload)
    return GetCurrentSession().request(
        method,
        "/api/linux/forward",
        headers={"User-Agent": GetCurrentSession().UA_LINUX_API},
        data={**LinuxApiEncrypt(payload)},
    )


@BaseWrapper
@EapiEncipered
def EapiCryptoRequest(url, plain, method):
    """Eapi - 适用于 新版客户端绝大部分 APIs"""
    cookies = {
        **GetCurrentSession().CONFIG_EAPI,
        "requestId": f"{int(time() * 1000)}_0233",
        "__csrf": GetCurrentSession().csrf_token,
        # **GetCurrentSession().cookies,
    }
    payload = {**plain, "header": json.dumps(cookies)}
    request = GetCurrentSession().request(
        method,
        url,
        headers={"User-Agent": GetCurrentSession().UA_EAPI, "Referer": None},
        cookies=cookies,
        data={
            **EapiEncrypt(
                parse(url).path.replace("/eapi/", "/api/"), json.dumps(payload)
            )
        },
    )
    return request.content


# endregion
# endregion
from . import (
    miniprograms,
    album,
    cloud,
    cloudsearch,
    login,
    playlist,
    track,
    user,
    video,
)



# -*- coding: utf-8 -*-
"""歌单 - Playlist APIs"""
from . import EapiCryptoRequest, LapiCryptoRequest, WeapiCryptoRequest
import json


@WeapiCryptoRequest
def GetPlaylistInfo(playlist_id, offset=0, total=True, limit=1000):
    """网页端 - 获取歌单内容

    Args:
        playlist_id ([type]): 歌单 ID
        offset (int, optional): 获取偏移数. Defaults to 0.
        total (bool, optional): 是否获取全部内容. Defaults to True.
        limit (int, optional): 单次获取量. Defaults to 30.

    Returns:
        dict
    """
    return "/weapi/v6/playlist/detail", {
        "id": str(playlist_id),
        "offset": str(offset),
        "total": str(total).lower(),
        "limit": str(limit),
        "n": str(limit),
    }


@WeapiCryptoRequest
def GetPlaylistComments(playlist_id: str, offset=0, limit=20, beforeTime=0):
    """网页端 - 获取歌单评论

    Args:
        album_id (str): 歌单ID
        offset (int, optional): 时间顺序偏移数. Defaults to 0.
        limit (int, optional): 单次评论获取量. Defaults to 20.
        beforeTime (int, optional): 评论将从该时间戳（秒为单位）开始. Defaults to 0.

    Returns:
        dict
    """
    return "/v1/resource/comments/A_PL_0_%s" % playlist_id, {
        "rid": str(playlist_id),
        "limit": str(limit),
        "offset": str(offset),
        "beforeTime": str(beforeTime * 1000),
    }


@LapiCryptoRequest
def GetTopPlaylists():
    """Linux PC端 - 获取热门歌单

    Returns:
        dict
    """
    return "/api/toplist", {}


@WeapiCryptoRequest
def SetManipulatePlaylistTracks(trackIds, playlistId, op="add", imme=True, e_r=True):
    """PC 端 - 操作歌单

    - op 有以下几种：
     - `add` : 添加到歌单
     - `del` : 从歌单删除

    Args:
        trackIds (list[str]): 待操作的歌曲ID列表
        playlistId ([int]): 要操作的歌单ID
        op (str, optional): 操作 . Defaults to "add".
        imme (bool, optional): 暂存. Defaults to True.

    Returns:
        [type]: [description]
    """
    trackIds = trackIds if isinstance(trackIds, list) else [trackIds]
    return "/weapi/v1/playlist/manipulate/tracks", {
        "trackIds": json.dumps(trackIds),
        "pid": str(playlistId),
        "op": op,
        "imme": str(imme).lower(),
    }


@EapiCryptoRequest
def SetCreatePlaylist(name: str, privacy=False):
    """PC 端 - 新建歌单

    Args:
        name (str): 歌单名
        privacy (bool, optional) : 是否私享. Defaults to False.
    """
    return "/eapi/playlist/create", {
        "name": str(name),
        "privacy": str(privacy * 1),        
    }


@EapiCryptoRequest
def SetRemovePlaylist(ids: list, self=True):
    """移动端 - 删除歌单

    Args:
        ids (list): 歌单 ID
        self (bool): 未知. Defaults to True
    Returns:
        dict
    """
    ids = ids if isinstance(ids, list) else [ids]
    return "/eapi/playlist/remove", {
        "ids": str(ids),
        "self": str(self),        
    }



# -*- coding: utf-8 -*-
"""登录、Anti - AntiCSRF 有关 APIs"""
from . import (
    EapiCryptoRequest,
    WeapiCryptoRequest,
    GetCurrentSession,
    logger,
    LoginFailedException,
)
from ..utils.crypto import HashHexDigest
import time


def WriteLoginInfo(response):
    """写登录态入当前 Session

    Args:
        response (dict): 解码后的登录态

    Raises:
        LoginFailedException: 登陆失败时发生
    """
    sess = GetCurrentSession()
    sess.login_info = {"tick": time.time(), "content": response}
    if not sess.login_info["content"]["code"] == 200:
        sess.login_info["success"] = False
        raise LoginFailedException(sess.login_info["content"])
    sess.login_info["success"] = True
    cookie = sess.cookies.get_dict()
    sess.csrf_token = cookie["__csrf"]
    logger.debug("Updated login info for user %s" % sess.nickname)


@WeapiCryptoRequest
def LoginLogout():
    """网页端 - 登出账号

    Returns:
        dict
    """
    return "/weapi/logout", {}


@WeapiCryptoRequest
def LoginRefreshToken():
    """网页端 - 刷新登录令牌

    Returns:
        dict
    """
    return "/weapi/w/login/cellphone", {}


@WeapiCryptoRequest
def LoginQrcodeUnikey(type=1):
    """网页端 - 获取二维码登录令牌

    - 该令牌UUID适用于以下 URL：
     - music.163.com/login?codekey={UUID}
    - 若需要使用该URL，应由网易云音乐移动客户端扫描其二维码 - 链接不会触发入口
    - 登录态将在 `LoginQrcodeCheck` 中更新，需周期性检测

    Args:
        type (int, optional): 未知. Defaults to 1.

    Returns:
        dict
    """
    return "/weapi/login/qrcode/unikey", {"type": str(type)}


@WeapiCryptoRequest
def LoginQrcodeCheck(unikey, type=1):
    """网页端 - 二维码登录状态检测

    Args:
        key (str): 二维码 unikey
        type (int, optional): 未知. Defaults to 1.

    Returns:
        dict
    """
    return "/weapi/login/qrcode/client/login", {"key": str(unikey), "type": type}


@WeapiCryptoRequest
def LoginTypeSwitch():
    """网页端 - 切换登录方式 - reserved

    Returns:
        dict
    """
    return "/weapi/w/user/login/type/switch", {}


@WeapiCryptoRequest
def GetCurrentLoginStatus():
    """网页端 - 获取当前登录态

    Returns:
        dict
    """
    return "/weapi/w/nuser/account/get", {}


def LoginViaCellphone(phone="", password="", ctcode=86, remeberLogin=True) -> dict:
    """网页端 - 手机号登陆

    Args:
        phone (str, optional): 手机号. Defaults to ''.
        password (str, optional): 明文密码. Defaults to ''.
        countrycode (int, optional): 国家代码. Defaults to 86.
        remeberLogin (bool, optional): 是否‘自动登录’，设置 `False` 可能导致权限问题. Defaults to True.

    Raises:
        LoginFailedException: 登陆失败时发生

    Returns:
        dict
    """
    path = "/eapi/login/cellphone"
    sess = GetCurrentSession()
    passwordHash = HashHexDigest(password)
    login_status = EapiCryptoRequest(
        lambda: (
            path,
            {
                "phone": str(phone),
                "password": str(passwordHash),
                "rememberLogin": str(remeberLogin).lower(),
                "countrycode": str(ctcode),
            },
        )
    )()
    WriteLoginInfo(login_status)
    return sess.login_info


@WeapiCryptoRequest
def SetSendRegisterVerifcationCodeViaCellphone(cell: str, ctcode=86):
    """网页端 - 发送验证码

    - 验证码 24h 内最多发送五次

    Args:
        cell (str): 手机号
        ctcode (int, optional): 国家代码. Defaults to 86.

    Returns:
        dict
    """
    return "/weapi/sms/captcha/sent", {"cellphone": str(cell), "ctcode": ctcode}


@WeapiCryptoRequest
def GetRegisterVerifcationStatusViaCellphone(cell: str, captcha: str, ctcode=86):
    """网页端 - 检查验证码是否正确

    Args:
        cell (str): 手机号
        captcha (str): 验证码
        ctcode (int, optional): 国家代码. Defaults to 86.

    Returns:
        dict
    """
    return "/weapi/sms/captcha/verify", {
        "cellphone": str(cell),
        "captcha": str(captcha),
        "ctcode": ctcode,
    }


@WeapiCryptoRequest
def SetRegisterAccountViaCellphone(
    cell: str, captcha: str, nickname: str, password: str
):
    """网页端 - 手机号注册

    - 需要已通过 `SetSendRegisterVerifcationCodeViaCellphone` 发送验证码
    - `忘记密码` 同样使用该 API
    - 成功后，现`Session`将登陆进入该账号

    Args:
        cell (str): 手机号
        captcha (str): 验证码
        nickname (str): 昵称
        password (str): 密码

    Returns:
        dict
    """
    return "/weapi/w/register/cellphone", {
        "captcha": str(captcha),
        "nickname": str(nickname),
        "password": HashHexDigest(password),
        "phone": str(cell),        
    }


@EapiCryptoRequest
def CheckIsCellphoneRegistered(cell: str, prefix=86):
    """移动端 - 检查某手机号是否已注册

    Args:
        cell (str): 手机号
        prefix (int): 区号 . Defaults to 86

    Returns:
        dict
    """
    return "/eapi/cellphone/existence/check", {"cellphone": cell, "countrycode": prefix}



# -*- coding: utf-8 -*-
# Helper utilites to aid operations w/ API responses
import logging, time

truncate_length = 64
logger = logging.getLogger("NCMHelper")


def TrackHelperProperty(default=None):
    def preWrapper(func):
        @property
        def wrapper(*a, **k):
            try:
                return func(*a, **k)
            except:
                logger.warn(
                    "Error while getting track attribute %s,using fallback value %s"
                    % (func.__name__, default)
                )
                return default

        return wrapper

    return preWrapper


class TrackHelper:
    """Helper class for handling generic track objects"""

    def __init__(self, track_dict) -> None:
        self.track = track_dict

    @TrackHelperProperty()
    def ID(self):
        """Track ID"""
        return self.track["id"]

    @TrackHelperProperty()
    def TrackPublishTime(self):
        """The publish year of the track"""
        epoch = (
            self.track["publishTime"] / 1000
        )  # stored as unix timestamp though only the year was ever useful
        return time.gmtime(epoch).tm_year

    @TrackHelperProperty()
    def TrackNumber(self):
        """The # of the track"""
        return self.track["no"]

    @TrackHelperProperty(default="Unknown")
    def TrackName(self):
        """The name of the track"""
        assert self.track["name"] != None
        return self.track["name"]

    @TrackHelperProperty(default="Unknown")
    def AlbumName(self):
        """The name of the album"""
        if self.track["al"]["id"]:
            return self.track["al"]["name"]
        else:
            return self.track["pc"]["alb"]

    @TrackHelperProperty(
        default="https://p1.music.126.net/UeTuwE7pvjBpypWLudqukA==/3132508627578625.jpg"
    )
    def AlbumCover(self):
        """The cover URL of the track's album"""
        al = self.track["al"] if "al" in self.track.keys() else self.track["album"]
        if al["id"]:
            return al["picUrl"]
        else:
            return "https://music.163.com/api/img/blur/" + self.track["pc"]["cid"]
            # source:PC version's core.js

    @TrackHelperProperty(default=["Various Artists"])
    def Artists(self):
        """All the artists' names as a list"""
        ar = self.track["ar"] if "ar" in self.track.keys() else self.track["artists"]
        ret = [_ar["name"] for _ar in ar]
        if not ret.count(None):
            return ret
        else:
            return [self.track["pc"]["ar"]]  # for NCM cloud-drive stored audio

    @TrackHelperProperty()
    def Title(self):
        """A formatted title for this song : {TrackName} - {ArtistsName}"""
        return f'{self.TrackName} - {",".join(self.Artists)}'

    _illegal_chars = set('\x00\\/:*?"<>|')

    @TrackHelperProperty()
    def SanitizedTitle(self):
        """Sanitized title for FS compatibility; Limits max length to 200 chars,
        and substitutes illegal chars with their full-width counterparts"""

        def _T(s, l=100):
            return "".join(
                [c if not c in self._illegal_chars else chr(ord(c) + 0xFEE0) for c in s]
            )[:l]

        return f'{_T(self.TrackName,100)} - {_T(",".join(self.Artists),100)}'



# -*- coding: utf-8 -*-
"""专辑 - Album APIs"""
from . import WeapiCryptoRequest


@WeapiCryptoRequest
def GetAlbumInfo(album_id: str):
    """网页端 - 获取专辑信息

    Args:
        album_id (str): 专辑ID

    Returns:
        dict
    """
    return "/weapi/v1/album/%s" % album_id, {}


@WeapiCryptoRequest
def GetAlbumComments(album_id: str, offset=0, limit=20, beforeTime=0):
    """网页端 - 获取专辑评论

    Args:
        album_id (str): 专辑ID
        offset (int, optional): 时间顺序偏移数. Defaults to 0.
        limit (int, optional): 单次评论获取量. Defaults to 20.
        beforeTime (int, optional): 评论将从该时间戳（秒为单位）开始. Defaults to 0.

    Returns:
        dict
    """
    return "/weapi/v1/resource/comments/R_AL_3_%s" % album_id, {
        "rid": "R_AL_3_%s" % album_id,
        "offset": str(offset),
        "total": "true",
        "limit": str(limit),
        "beforeTime": str(beforeTime * 1000),
    }



# -*- coding: utf-8 -*-
# PyNCM CLI interface

from pyncm import (
    DumpSessionAsString,
    GetCurrentSession,
    LoadSessionFromString,
    SetCurrentSession,
    __version__,
    logger,
)
from pyncm.utils.lrcparser import LrcParser
from pyncm.utils.helper import TrackHelper
from pyncm.apis import login, track, playlist, album

from queue import Queue
from concurrent.futures import ThreadPoolExecutor
from threading import Thread
from time import sleep
from os.path import join, exists
from os import remove, makedirs

from logging import getLogger, basicConfig
import sys, argparse, re

# Import checks
OPTIONALS = {"mutagen": False, "tqdm": False}
OPTIONALS_MISSING_INFO = {"mutagen": "无法为下载的音乐添加歌手信息，封面等资源", "tqdm": "将不会显示下载进度条"}
from importlib.util import find_spec

for import_name in OPTIONALS:
    OPTIONALS[import_name] = find_spec(import_name)
    if not OPTIONALS[import_name]:
        sys.stderr.writelines(
            [f"[WARN] {import_name} 没有安装，{OPTIONALS_MISSING_INFO[import_name]}\n"]
        )

__desc__ = """PyNCM 网易云音乐下载工具 %s""" % __version__
max_workers = 4

BITRATES = {"standard": 96000, "high": 320000, "lossless": 3200000}
# Key-Value classes
class BaseKeyValueClass:
    def __init__(self, **kw) -> None:
        for k, v in kw.items():
            self.__setattr__(k, v)


class BaseDownloadTask(BaseKeyValueClass):
    id: int
    url: str
    dest: str


class LyricsDownloadTask(BaseDownloadTask):
    id: int
    dest: str
    lrc_blacklist: set


class TrackDownloadTask(BaseKeyValueClass):
    song: TrackHelper
    cover: BaseDownloadTask
    lyrics: BaseDownloadTask
    resource: BaseDownloadTask
    resource_info: dict
    lyrics_exclude: set
    save_as: str


class TaskPoolExecutorThread(Thread):
    @staticmethod
    def tag_audio(track: TrackHelper, file: str, cover_img: str = ""):
        if not OPTIONALS["mutagen"]:
            return

        def write_keys(song):
            # Write trackdatas
            song["title"] = track.TrackName
            song["artist"] = track.Artists
            song["album"] = track.AlbumName
            song["tracknumber"] = str(track.TrackNumber)
            song["date"] = str(track.TrackPublishTime)
            song.save()

        def mp4():
            from mutagen import easymp4
            from mutagen.mp4 import MP4, MP4Cover

            song = easymp4.EasyMP4(file)
            write_keys(song)
            if exists(cover_img):
                song = MP4(file)
                song["covr"] = [MP4Cover(open(cover_img, "rb").read())]
                song.save()

        def mp3():
            from mutagen.mp3 import EasyMP3
            from mutagen.id3 import ID3, APIC

            song = EasyMP3(file)
            write_keys(song)
            if exists(cover_img):
                song = ID3(file)
                song.update_to_v23()  # better compatibility over v2.4
                song.add(
                    APIC(
                        encoding=3,
                        mime="image/jpeg",
                        type=3,
                        desc="",
                        data=open(cover_img, "rb").read(),
                    )
                )
                song.save(v2_version=3)

        def flac():
            from mutagen.flac import FLAC, Picture

            song = FLAC(file)
            write_keys(song)
            if exists(cover_img):
                pic = Picture()
                pic.data = open(cover_img, "rb").read()
                pic.mime = "image/jpeg"
                song.add_picture(pic)
                song.save()

        def ogg():
            import base64
            from mutagen.flac import Picture
            from mutagen.oggvorbis import OggVorbis

            song = OggVorbis(file)
            write_keys(song)
            if exists(cover_img):
                pic = Picture()
                pic.data = open(cover_img, "rb").read()
                pic.mime = "image/jpeg"
                song["metadata_block_picture"] = [
                    base64.b64encode(pic.write()).decode("ascii")
                ]
                song.save()

        format = file.split(".")[-1].upper()
        for ext, method in [
            ({"M4A", "M4B", "M4P", "MP4"}, mp4),
            ({"MP3"}, mp3),
            ({"FLAC"}, flac),
            ({"OGG", "OGV"}, ogg),
        ]:
            if format in ext:
                return method() or True
        return False

    def download_by_url(self, url, dest, xfer=False):
        # Downloads generic content
        response = GetCurrentSession().get(url, stream=True)
        length = int(response.headers.get("content-length"))

        with open(dest, "wb") as f:
            for chunk in response.iter_content(128 * 2**10):
                self.xfered += len(chunk)
                if xfer:
                    self.finished_tasks += len(chunk) / length  # task [0,1]
                f.write(chunk)  # write every 128KB read
        return dest

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.finished_tasks: float = 0
        self.xfered = 0
        self.task_queue = Queue()

    def run(self):
        def execute(task: TrackDownloadTask):
            try:
                if not exists(task.resource.dest):
                    makedirs(task.resource.dest)
                # Downloding source audio
                dest_src = self.download_by_url(
                    task.resource.url,
                    join(
                        task.resource.dest,
                        task.save_as + "." + task.resource_info["type"],
                    ),
                    xfer=True,
                )
                # Downloading cover
                dest_cvr = self.download_by_url(
                    task.cover.url, join(task.cover.dest, "%s.jpg" % task.cover.id)
                )
                # Downloading & Parsing lyrics
                dest_lrc = join(task.lyrics.dest, task.save_as + ".lrc")
                lrc = LrcParser()
                dLyrics = track.GetTrackLyrics(task.lyrics.id)
                for k in set(dLyrics.keys()) & (
                    {"lrc", "tlyric", "romalrc"} - task.lyrics.lrc_blacklist
                ):  # Filtering LRCs
                    lrc.LoadLrc(dLyrics[k]["lyric"])
                lrc_text = lrc.DumpLyrics()
                if lrc_text:
                    open(dest_lrc, "w", encoding="utf-8").write(lrc_text)
                # Tagging the audio
                self.tag_audio(task.song, dest_src, dest_cvr)
                # Cleaning up
                remove(dest_cvr)
            except Exception as e:
                logger.warning("下载未完成 %s - %s" % (task.song.Title, e))

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            while True:
                task = self.task_queue.get()
                future = executor.submit(execute, task)
                future.add_done_callback(lambda future: self.task_queue.task_done())


# Subroutines
class Subroutine:
    """Generic subroutine

    Subroutines are `callable`,upon called with `ids`,one
    queues tasks with all given arguments via `put_func` callback
    """

    def __init__(self, args, put_func) -> None:
        self.args = args
        self.put = put_func


def create_subroutine(sub_type) -> Subroutine:
    """Dynamically creates subroutine callable by string specified"""

    class Playlist(Subroutine):
        def forIds(self, ids):
            dDetails = track.GetTrackDetail(ids)["songs"]
            dAudios = track.GetTrackAudio(ids, bitrate=BITRATES[self.args.quality])[
                "data"
            ]

            dDetails = sorted(dDetails, key=lambda song: song["id"])
            dAudios = sorted(dAudios, key=lambda song: song["id"])            
            for index, (dDetail, dAudio) in enumerate(zip(dDetails, dAudios)):
                try:
                    song = TrackHelper(dDetail)
                    if dAudio["type"]:
                        logger.info(
                            "单曲 #%d / %d - %s - %s (%dkbps) - %s"
                            % (
                                index + 1,
                                len(dDetails),
                                song.Title,
                                song.AlbumName,
                                dAudio["br"] // 1000,
                                str(dAudio["type"]).upper(),
                            )
                        )
                        tSong = TrackDownloadTask(
                            song=song,
                            cover=BaseDownloadTask(
                                id=song.ID, url=song.AlbumCover, dest=self.args.output
                            ),
                            resource=BaseDownloadTask(
                                id=song.ID, url=dAudio["url"], dest=self.args.output
                            ),
                            lyrics=LyricsDownloadTask(
                                id=song.ID,
                                dest=self.args.output,
                                lrc_blacklist=set(self.args.lyric_no),
                            ),
                            resource_info=dAudio,
                            save_as=self.args.template.format(
                                **{
                                    "id": song.ID,
                                    "year": song.TrackPublishTime,
                                    "no": song.TrackNumber,
                                    "track": song.TrackName,
                                    "album": song.AlbumName,
                                    "title": song.SanitizedTitle,
                                    "artists": " / ".join(song.Artists),
                                }
                            ),
                        )
                        self.put(tSong)                        
                    else:
                        logger.warn(
                            "单曲 #%d / %d - %s - %s 资源不存在"
                            % (
                                index + 1,
                                len(dDetails),
                                song.Title,
                                song.AlbumName,                                
                            )
                        )
                except Exception as e:
                    logger.warning(
                        "单曲 #%d / %d - %s - %s 无法下载： %s"
                        % (index + 1, len(dDetails), song.Title, song.AlbumName, e)
                    )
            return index

        def __call__(self, ids):
            queued = 0
            for _id in ids:
                dList = playlist.GetPlaylistInfo(_id)
                logger.info("歌单 ：%s" % dList["playlist"]["name"])
                queued += self.forIds(
                    [tid["id"] for tid in dList["playlist"]["trackIds"]]
                )
            return queued

    class Album(Playlist):
        def __call__(self, ids):
            queued = 0
            for _id in ids:
                dList = album.GetAlbumInfo(_id)
                logger.info("专辑 ：%s" % dList["album"]["name"])
                queued += self.forIds([tid["id"] for tid in dList["songs"]])
            return queued

    class Song(Playlist):
        def __call__(self, ids):
            return self.forIds(ids)

    return {"song": Song, "playlist": Playlist, "album": Album}[sub_type]


def parse_sharelink(url):
    """Parses (partial) URLs for NE resources and determines its ID and type

    e.g.
        31140560 (plain song id)
        https://greats3an.github.io/pyncmd/?trackId=1818064296 (pyncmd)
        分享Ali Edwards的单曲《Devil Trigger》: http://music.163.com/song/1353163404/?userid=6483697162 (来自@网易云音乐) (mobile app)
        "分享mos9527创建的歌单「東方 PC」: http://music.163.com/playlist?id=72897851187" (desktop app)
    """
    rurl = re.findall("(?:http|https):\/\/.*", url)
    if rurl:
        url = rurl[0]  # Use first URL found. Otherwise use value given as is.
    numerics = re.findall("\d{4,}", url)
    assert numerics, "未在链接中找到任何 ID"
    ids = numerics[:1]  # Only pick the first match
    table = {
        "song": ["trackId", "song"],
        "playlist": ["playlist"],
        "album": ["album"],
    }
    rtype = "song"  # Defaults to songs (tracks)
    for rtype_, rkeyword in table.items():
        for kw in rkeyword:
            if kw in url:
                rtype = rtype_
                break  # Match type by keyword
    return rtype, ids


def parse_args():
    """Setting up __main__ argparser"""
    parser = argparse.ArgumentParser(
        description=__desc__, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument("url", metavar="链接", help="网易云音乐分享链接")
    group = parser.add_argument_group("下载")
    group.add_argument(
        "--template",
        metavar="模板",
        help=r"""保存文件名模板
    参数：    
        id     - 网易云音乐资源 ID
        year   - 出版年份
        no     - 专辑中编号
        album  - 专辑标题
        track  - 单曲标题        
        title  - 完整标题
        artists- 艺术家名
    例：
        {track} - {artists} 等效于 {title}""",
        default=r"{title}",
    )
    group.add_argument(
        "--quality",
        metavar="音质",
        choices=list(BITRATES.keys()),
        help=r"""音频音质（高音质需要 CVIP）
    参数：
        lossless - “无损”
        high     - 较高
        standard - 标准""",
        default="standard",
    )
    group.add_argument("--output", metavar="输出", default=".", help="输出文件夹")
    group = parser.add_argument_group("歌词")
    group.add_argument(
        "--lyric-no",
        metavar="跳过歌词",
        help=r"""跳过某些歌词类型的合并
    参数：
        lrc    - 源语言歌词
        tlyric - 翻译后歌词
        romalrc- 罗马音歌词
    例：
        --lyric-no tlyric --lyric-no romalrc 将只下载源语言歌词""",
        choices=["lrc", "tlyric", "romalrc"],
        default="",
        nargs="+",
    )
    group = parser.add_argument_group("登陆")
    group.add_argument("--phone", metavar="手机", default="", help="网易账户手机号")
    group.add_argument("--pwd",'--password',metavar="密码", default="", help="网易账户密码")
    group.add_argument("--save", metavar="[保存到]", default="", help="写本次登录信息于文件")
    group.add_argument(
        "--load", metavar="[保存的登陆信息文件]", default="", help="从文件读取登录信息供本次登陆使用"
    )
    group.add_argument("--http",action='store_true',help='优先使用 HTTP，不保证不被升级')
    group.add_argument("--log-level",help='日志等级',default='INFO')

    args = parser.parse_args()
    rtype, ids = parse_sharelink(args.url)

    return ids, args, rtype


def __main__():
    ids, args, rtype = parse_args()
    basicConfig(level=args.log_level, format="[%(levelname).4s] %(message)s")

    if args.load:
        logger.info("读取登录信息 : %s" % args.load)
        SetCurrentSession(LoadSessionFromString(open(args.load).read()))
    if args.http:
        GetCurrentSession().force_http = True
        logger.warning('优先使用 HTTP')
    if args.phone and args.pwd:
        login.LoginViaCellphone(args.phone, args.pwd)
        logger.info(
            "账号 ：%s (VIP %s)"
            % (GetCurrentSession().nickname, GetCurrentSession().vipType)
        )
    if args.save:
        logger.info("保存登陆信息于 : %s" % args.save)
        open(args.save, "w").write(DumpSessionAsString(GetCurrentSession()))

    executor = TaskPoolExecutorThread()
    executor.daemon = True
    executor.start()

    def enqueue_task(task):
        executor.task_queue.put(task)

    subroutine = create_subroutine(rtype)(args, enqueue_task)
    total_queued = subroutine(ids)  # Enqueue tasks

    if OPTIONALS["tqdm"]:
        import tqdm

        _tqdm = tqdm.tqdm(
            bar_format="已完成 {desc}: {percentage:.1f}%|{bar}| {n:.2f}/{total_fmt} [{elapsed}<{remaining}"
        )
        _tqdm.total = total_queued

        def report():
            _tqdm.desc = _tqdm.format_sizeof(executor.xfered, suffix="B", divisor=1024)
            _tqdm.update(min(executor.finished_tasks, total_queued) - _tqdm.n)
            return True

    else:

        def report():
            sys.stderr.write(
                f"下载中 : {executor.finished_tasks:.1f} / {total_queued} ({(executor.finished_tasks * 100 / total_queued):.1f} %,{executor.xfered >> 20} MB)               \r"
            )
            return True

    while executor.task_queue.unfinished_tasks >= 0:
        report() and sleep(0.5)
        if executor.task_queue.unfinished_tasks == 0:
            break

    return 0


if __name__ == "__main__":
    logger = getLogger()    
    sys.exit(__main__())
